@@ -2,7 +2,7 @@
  *
  * This file is part of XDelta - A binary delta generator.
  *
- * Copyright (C) 1997, 1998, 1999  Josh MacDonald
+ * Copyright (C) 1997, 1998, 1999, 2001  Josh MacDonald
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
  *
  * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
  *
- * $Id: xdelta.c 1.4.1.50 Fri, 09 Apr 1999 15:24:35 -0700 jmacd $
+ * $Id: xdelta.c 1.4.1.50.1.2 Fri, 29 Jun 2001 06:01:08 -0700 jmacd $
  */
 
 #define G_DISABLE_ASSERT
@@ -37,7 +37,7 @@ const guint xdelta_major_version = 1;
 /* $Format: "const guint xdelta_minor_version = $ReleaseMinorVersion$;" $ */
 const guint xdelta_minor_version = 1;
 /* $Format: "const guint xdelta_micro_version = $ReleaseMicroVersion$;" $ */
-const guint xdelta_micro_version = 1;
+const guint xdelta_micro_version = 3;
 
 /* Control functions.
  */
@@ -46,6 +46,65 @@ static XdeltaControl* control_version_0 (SerialVersion0Control* cont) ;
 static void           control_copy   (XdeltaControl* cont, XdeltaSource* src, guint from, guint to);
 static gboolean       control_add_info (XdeltaControl* cont, XdeltaSource* src, const guint8* md5, guint len);
 
+#ifndef XDELTA_HARDCODE_SIZE
+int QUERY_SIZE = 0;
+int QUERY_SIZE_POW = 0;
+int QUERY_SIZE_MASK = 0;
+#endif
+
+int xdp_set_query_size_pow (int size_pow)
+{
+#ifdef XDELTA_HARDCODE_SIZE
+  return XDP_QUERY_HARDCODED;
+#else
+
+  int x        = 1;
+  int size_log = 0;
+
+  for (; x != 0; x <<= 1, size_log += 1)
+    {
+      if (x == size_pow)
+	goto good;
+    }
+
+  return XDP_QUERY_POW2;
+
+ good:
+
+  QUERY_SIZE      = size_log;
+  QUERY_SIZE_POW  = size_pow;
+  QUERY_SIZE_MASK = size_pow-1;
+
+  return 0;
+
+#endif
+}
+
+int
+xdp_blocksize ()
+{
+  if (QUERY_SIZE == 0)
+    {
+      xdp_set_query_size_pow (1<<QUERY_SIZE_DEFAULT);
+    }
+
+  return QUERY_SIZE_POW;
+}
+
+const char*
+xdp_errno (int errval)
+{
+  switch (errval)
+    {
+    case XDP_QUERY_HARDCODED:
+      return "query size hardcoded";
+    case XDP_QUERY_POW2:
+      return "query size must be a power of 2";
+    }
+
+  return g_strerror (errval);
+}
+
 const guint16 single_hash[256] =
 {
   /* Random numbers generated using SLIB's pseudo-random number
@@ -130,7 +189,8 @@ generate_checksums (XdeltaStream    *stream,
   if (total_checksums == 0)
     return TRUE;
 
-  result = g_new (XdeltaChecksum, total_checksums);
+  /* This is the in-core FROM checksum table. */
+  result         = g_new (XdeltaChecksum, total_checksums);
   source->cksums = result;
 
   for (pages = handle_pages (stream); segment_page <= pages; segment_page += 1)
@@ -886,6 +946,11 @@ xdp_generate_delta_int (XdeltaGenerator *gen,
   gint total_from_len = 0;
   guint32* table = NULL;
 
+  if (QUERY_SIZE == 0)
+    {
+      xdp_set_query_size_pow (1<<QUERY_SIZE_DEFAULT);
+    }
+
   if (gen->sources->len == 0)
     {
       xd_generate_void_event (EC_XdTooFewSources);
