@@ -2,7 +2,7 @@
  *
  * This file is part of XDelta - A binary delta generator.
  *
- * Copyright (C) 1997, 1998, 1999  Josh MacDonald
+ * Copyright (C) 1997, 1998, 1999, 2001  Josh MacDonald
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
  *
  * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
  *
- * $Id: xdmain.c 1.22 Fri, 09 Apr 1999 15:24:35 -0700 jmacd $
+ * $Id: xdmain.c 1.22.1.5 Fri, 29 Jun 2001 06:01:08 -0700 jmacd $
  */
 
 #include <stdio.h>
@@ -32,18 +32,33 @@
 #include <fcntl.h>
 #include <sys/stat.h>
 
-#ifndef _WIN32
+#if defined(_WIN32) || defined(__DJGPP__)
+#define WINHACK
+#endif
+
+#ifndef WINHACK
 #include <unistd.h>
 #include <sys/mman.h>
 #define O_BINARY 0
-#else
+
+#else /* WINHACK */
+
 #include <io.h>
 #include <process.h>
+#ifdef __DJGPP__
+#include <unistd.h>
+#include <dpmi.h>
+#endif
+
 #define STDOUT_FILENO 1
 #define lstat stat
+
+#ifndef __DJGPP__
 #define S_IFMT _S_IFMT
 #define S_IFREG _S_IFREG
-#endif
+#endif /* !__DJGPP__ */
+
+#endif /* !WINHACK_ */
 
 #include <zlib.h>
 
@@ -80,7 +95,7 @@ extern HandleFuncTable xd_handle_table;
 /* and, the header is follwed by the from file name, then the to file
  * name, then the data. */
 
-#ifdef _WIN32
+#ifdef WINHACK
 #define FOPEN_READ_ARG "rb"
 #define FOPEN_WRITE_ARG "wb"
 #define FILE_SEPARATOR '\\'
@@ -88,7 +103,7 @@ extern HandleFuncTable xd_handle_table;
 #define FOPEN_READ_ARG "r"
 #define FOPEN_WRITE_ARG "w"
 #define FILE_SEPARATOR '/'
-#endif
+#endif /* WINHACK */
 
 #include "getopt.h"
 
@@ -178,7 +193,7 @@ struct _XdFileHandle
 };
 
 /* $Format: "static const char xdelta_version[] = \"$ReleaseVersion$\"; " $ */
-static const char xdelta_version[] = "1.1.1"; 
+static const char xdelta_version[] = "1.1.3"; 
 
 typedef struct _Command Command;
 
@@ -208,9 +223,8 @@ static struct option const long_options[] =
   {"noverify",            no_argument, 0, 'n'},
   {"pristine",            no_argument, 0, 'p'},
   {"quiet",               no_argument, 0, 'q'},
-  /*{"long-long-format",    no_argument, 0, 'L'},
-    {"long-format",         no_argument, 0, 'l'},*/
   {"maxmem",              required_argument, 0, 'm'},
+  {"blocksize",           required_argument, 0, 's'},
   {0,0,0,0}
 };
 
@@ -221,8 +235,6 @@ static gint         pristine = FALSE;
 static gint         verbose = FALSE;
 static gint         max_mapped_pages = G_MAXINT;
 static gint         quiet = FALSE;
-/*static gint         long_format = FALSE;
-static gint         really_long_format = FALSE;*/
 
 #define xd_error g_warning
 
@@ -237,20 +249,21 @@ usage ()
 static void
 help ()
 {
-  xd_error ("usage: %s COMMAND [OPTIONS] [ARG1 ARG2 ...]\n", program_name);
+  xd_error ("usage: %s COMMAND [OPTIONS] [ARG1 ...]\n", program_name);
   xd_error ("COMMAND is one of:\n");
   xd_error ("  delta     Produce a delta from ARG1 to ARG2 producing ARG3\n");
   xd_error ("  info      List details about delta ARG1\n");
-  xd_error ("  patch     Patch file ARG2 with ARG1 producing ARG3\n");
+  xd_error ("  patch     Apply patch ARG1 using file ARG2 producing ARG3\n");
   xd_error ("OPTIONS are:\n");
-  /*xd_error ("  -b, --base64\n"); ADD THIS STUFF BACK */
-  xd_error ("  -v, --version\n");
-  xd_error ("  -V, --verbose\n");
-  xd_error ("  -h, --help\n");
-  xd_error ("  -n, --noverify\n");
-  xd_error ("  -p, --pristine\n");
+  xd_error ("  -v, --version      Print version information\n");
+  xd_error ("  -V, --verbose      Print verbose error messages\n");
+  xd_error ("  -h, --help         Print this summary\n");
+  xd_error ("  -n, --noverify     Disable automatic MD5 verification\n");
+  xd_error ("  -p, --pristine     Disable automatic GZIP decompression\n");
   xd_error ("  -m, --maxmem=SIZE  Set the buffer size limit, e.g. 640K, 16M\n");
-  xd_error ("  -[0-9]    Compression level: 0=none, 1=fast, 6=default, 9=best\n");
+  xd_error ("  -[0-9]             ZLIB compress level: 0=none, 1=fast, 6=default, 9=best\n");
+  xd_error ("  -s=BLOCK_SIZE      Sets block size (power of 2), minimum match length\n");
+  xd_error ("                     In-core memory requirement is (FROM_LEN * 8) / BLOCK_SIZE\n");
   exit (2);
 }
 
@@ -316,6 +329,21 @@ main (gint argc, gchar** argv)
   if (! xd_edsio_init ())
     return 2;
 
+#ifdef __DJGPP__
+  /*
+   * (richdawe@bigfoot.com): Limit maximum memory usage to 87.5% of available
+   * physical memory for DJGPP. Otherwise the replacement for mmap() will
+   * exhaust memory and XDelta will fail.
+   */
+  {
+    unsigned long phys_free = _go32_dpmi_remaining_physical_memory();
+
+    max_mapped_pages  = phys_free / XD_PAGE_SIZE;
+    max_mapped_pages %= 8;
+    max_mapped_pages *= 7;
+  }
+#endif /* DJGPP */
+
   program_name = g_basename (argv[0]);
 
   g_log_set_handler (G_LOG_DOMAIN,
@@ -349,18 +377,27 @@ main (gint argc, gchar** argv)
 
   while ((c = getopt_long(argc,
 			  argv,
-			  "+nqphvVm:0123456789",
+			  "+nqphvVs:m:0123456789",
 			  long_options,
 			  &longind)) != EOF)
     {
       switch (c)
 	{
-	  /*case 'l': long_format = TRUE; break;
-	    case 'L': really_long_format = TRUE; break;*/
 	case 'q': quiet = TRUE; break;
 	case 'n': no_verify = TRUE; break;
 	case 'p': pristine = TRUE; break;
 	case 'V': verbose = TRUE; break;
+	case 's':
+	  {
+	    int ret;
+	    int s = atoi (optarg);
+
+	    if ((ret = xdp_set_query_size_pow (s)) && ! quiet)
+	      {
+		xd_error ("illegal query size: %s\n", xdp_errno (ret));
+	      }
+	  }
+	  break;
 	case 'm':
 	  {
 	    gchar* end = NULL;
@@ -376,7 +413,15 @@ main (gint argc, gchar** argv)
 		return 2;
 	      }
 
+#ifdef __DJGPP__
+	    /*
+	     * (richdawe@bigfoot.com): Old MS-DOS systems may have a maximum
+	     * of 8MB memory == XD_PAGE_SIZE * 8. Therefore, do what the user
+	     * asks for on MS-DOS.
+	     */
+#else /* !__DJGPP__ */
 	    l = MAX (l, XD_PAGE_SIZE * 8);
+#endif /* __DJGPP__ */
 
 	    max_mapped_pages = l / XD_PAGE_SIZE;
 	  }
@@ -561,7 +606,11 @@ xd_tmpname (void)
 
   do
     {
-      g_string_sprintf (s, "%s/xdtmp.%d.%d", tmpdir, x, seq++);
+      /*
+       * (richdawe@bigfoot.com): Limit temporary filenames to
+       * the MS-DOS 8+3 convention for DJGPP.
+       */
+      g_string_sprintf (s, "%s/xd-%05d.%03d", tmpdir, x, seq++);
     }
   while (lstat (s->str, &buf) == 0);
 
@@ -588,7 +637,7 @@ file_gunzip (const char* name)
 
   if (nread < 0)
     {
-      xd_error ("gzread %s failed: %s\n", name, g_strerror (errno));
+      xd_error ("gzread %s failed\n", name);
       return NULL;
     }
 
@@ -657,11 +706,16 @@ open_read_noseek_handle (const char* name, gboolean* is_compressed, gboolean wil
 static void
 xd_read_close (XdFileHandle* fh)
 {
+  /*
+   * (richdawe@bigfoot.com): On Unix you can unlink a file while it is
+   * still open. On MS-DOS this can lead to filesystem corruption. Close
+   * the file before unlinking it.
+   */
+  close (fh->fd);
+
   if (fh->cleanup)
     unlink (fh->cleanup);
 
-  close (fh->fd);
-
   if (fh->in)
     (*fh->in_close) (fh);
 }
@@ -1101,7 +1155,7 @@ really_free_one_page (XdFileHandle* fh)
 
       if (to_unmap > 0)
 	{
-#ifdef WIN32
+#ifdef WINHACK
 	  g_free (lru_dead->buffer);
 #else
 	  if (munmap (lru_dead->buffer, to_unmap))
@@ -1109,7 +1163,7 @@ really_free_one_page (XdFileHandle* fh)
 	      xd_error ("munmap failed: %s\n", g_strerror (errno));
 	      return FALSE;
 	    }
-#endif
+#endif /* WINHACK */
 	}
 
       lru_dead->buffer = NULL;
@@ -1215,7 +1269,7 @@ xd_handle_map_page (XdFileHandle *fh, guint pgno, const guint8** mem)
 
       if (to_map > 0)
 	{
-#ifdef WIN32
+#ifdef WINHACK
 	  lru->buffer = g_malloc (to_map);
 
 	  if (lseek (fh->fd, pgno * XD_PAGE_SIZE, SEEK_SET) < 0)
@@ -1438,41 +1492,41 @@ delta_command (gint argc, gchar** argv)
   guint32 control_offset, header_offset;
   const char* from_name, *to_name;
   guint32 header_space[HEADER_WORDS];
+  int fd;
 
   memset (header_space, 0, sizeof (header_space));
 
-  /* This could read tofile from stdin... */
-  if (argc < 2 || argc > 3)
+  if (argc != 3)
     {
-      xd_error ("usage: %s delta fromfile tofile [patchfile]\n", program_name);
+      xd_error ("usage: %s delta fromfile tofile patchfile\n", program_name);
       return 2;
     }
 
+  if (verbose)
+    {
+      xd_error ("using block size: %d bytes\n", xdp_blocksize ());
+    }
+
+
   if (! (from = open_read_seek_handle (argv[0], &from_is_compressed, TRUE)))
     return 2;
 
   if (! (to = open_read_noseek_handle (argv[1], &to_is_compressed, FALSE, TRUE)))
     return 2;
 
-  if (argc == 2 || strcmp (argv[2], "-") == 0)
-    {
-      patch_out_fd = STDOUT_FILENO;
-      patch_out_name = "standard output";
-    }
-  else
-    {
-      int fd = open (argv[2], O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, 0666);
+  // Note: I tried support to patches to stdout, but it broke when
+  // compression was added.  Sigh
+  fd = open (argv[2], O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, 0666);
 
-      if (fd < 0)
-	{
-	  xd_error ("open %s failed: %s\n", argv[2], g_strerror (errno));
-	  return 2;
-	}
-
-      patch_out_fd = fd;
-      patch_out_name = argv[2];
+  if (fd < 0)
+    {
+      xd_error ("open %s failed: %s\n", argv[2], g_strerror (errno));
+      return 2;
     }
 
+  patch_out_fd = fd;
+  patch_out_name = argv[2];
+
   from_name = g_basename (argv[0]);
   to_name = g_basename (argv[1]);
 
@@ -1593,15 +1647,15 @@ process_patch (const char* name)
       patch->has_trailer = TRUE;
       patch->patch_version = "1.0.4";
     }
-  else if (strncmp (patch->magic_buf, XDELTA_100_PREFIX, XDELTA_PREFIX_LEN))
+  else if (strncmp (patch->magic_buf, XDELTA_100_PREFIX, XDELTA_PREFIX_LEN) == 0)
     {
       patch->patch_version = "1.0";
     }
-  else if (strncmp (patch->magic_buf, XDELTA_020_PREFIX, XDELTA_PREFIX_LEN))
+  else if (strncmp (patch->magic_buf, XDELTA_020_PREFIX, XDELTA_PREFIX_LEN) == 0)
     goto nosupport;
-  else if (strncmp (patch->magic_buf, XDELTA_018_PREFIX, XDELTA_PREFIX_LEN))
+  else if (strncmp (patch->magic_buf, XDELTA_018_PREFIX, XDELTA_PREFIX_LEN) == 0)
     goto nosupport;
-  else if (strncmp (patch->magic_buf, XDELTA_014_PREFIX, XDELTA_PREFIX_LEN))
+  else if (strncmp (patch->magic_buf, XDELTA_014_PREFIX, XDELTA_PREFIX_LEN) == 0)
     goto nosupport;
   else
     {
@@ -1745,7 +1799,7 @@ info_command (gint argc, gchar** argv)
 
   xd_error ("patch from segments: %d\n", patch->cont->source_info_len);
 
-  xd_error ("MD5\t\t\t\t Length\tCopies\tUsed\tSeq?\tName\n");
+  xd_error ("MD5\t\t\t\t\tLength\tCopies\tUsed\tSeq?\tName\n");
 
   for (i = 0; i < patch->cont->source_info_len; i += 1)
     {
@@ -1753,7 +1807,7 @@ info_command (gint argc, gchar** argv)
 
       edsio_md5_to_string (si->md5, buf);
 
-      xd_error ("%s %d\t%d\t%d\t%s\t%s\n",
+      xd_error ("%s\t%d\t%d\t%d\t%s\t%s\n",
 		buf,
 		si->len,
 		si->copies,
@@ -1771,10 +1825,12 @@ patch_command (gint argc, gchar** argv)
   XdFileHandle* to_out;
   XdeltaPatch* patch;
   gint to_out_fd;
+  int count = 1;
+  int ret;
 
   if (argc < 1 || argc > 3)
     {
-      xd_error ("usage: %s patch patchfile fromfile [tofile]\n", program_name);
+      xd_error ("usage: %s patch patchfile [fromfile [tofile]]\n", program_name);
       return 2;
     }
 
@@ -1784,12 +1840,25 @@ patch_command (gint argc, gchar** argv)
   if (argc > 1)
     patch->from_name = argv[1];
   else if (verbose)
-    xd_error ("using default from file name: %s\n", patch->from_name);
+    xd_error ("using from file name: %s\n", patch->from_name);
 
   if (argc > 2)
     patch->to_name = argv[2];
-  else if (verbose)
-    xd_error ("using default to file name: %s\n", patch->to_name);
+  else
+    {
+      struct stat sbuf;
+      gchar *defname = g_strdup (patch->to_name);
+
+      while ((ret = stat (patch->to_name, & sbuf)) == 0)
+	{
+	  if (verbose)
+	    xd_error ("to file exists: %s\n", patch->to_name);
+	  patch->to_name = g_strdup_printf ("%s.xdp%d", defname, count++);
+	}
+
+      if (verbose || strcmp (defname, patch->to_name) != 0)
+	xd_error ("using to file name: %s\n", patch->to_name);
+    }
 
   if (strcmp (patch->to_name, "-") == 0)
     {
@@ -1821,11 +1890,14 @@ patch_command (gint argc, gchar** argv)
 	return 2;
 
       if (from_is_compressed != ((patch->patch_flags & FLAG_FROM_COMPRESSED) && 1))
-	xd_error ("warning: expected %scompressed from file\n", (patch->patch_flags & FLAG_FROM_COMPRESSED) ? "" : "un");
+	xd_error ("warning: expected %scompressed from file (%s)\n",
+		  (patch->patch_flags & FLAG_FROM_COMPRESSED) ? "" : "un",
+		  patch->from_name);
 
       if (xd_handle_length (from_in) != patch->from_source->len)
 	{
-	  xd_error ("expected from file of %slength %d bytes\n",
+	  xd_error ("expected from file (%s) of %slength %d bytes\n",
+		    patch->from_name,
 		    from_is_compressed ? "uncompressed " : "",
 		    patch->from_source->len);
 	  return 2;
